/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#include "InternalFrame.h"
#include "InternalSideData.h"
#include <cstring>

namespace xma {

#define XMA_FRAME "Frame"
#define ROI_SB_INDEX 0

const char* pixel_format_names[] = {"none", "nv12", "nv21", "p010le", "p010be", "yuv420p", "yuv420p10le", "yuv420p10be", "yuv422p", "yuv422p10le", "yuv422p10be",
    "yuv444p", "argb", "abgr", "rgba", "bgra", "rgb24", "bgr24", "vpe", "packed10", "bgr0", "rgb16", "rgbp", "bgrp", "rgb48le", "bgr48le", "rgb48p", "bgr48p"};

//////////////////////////////////////////////////////////////////////////////
PrivateFrameInfo::PrivateFrameInfo(IntSessionFactoryPtr sf, LogPtr log, buffer_t buffer) : m_buffer(std::move(buffer)), m_sf(sf), m_log(log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateFrameInfo::PrivateFrameInfo()");
}

PrivateFrameInfo::~PrivateFrameInfo() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateFrameInfo::~PrivateFrameInfo()");
  if (m_pic) {
    m_log->LogMsg(XMA_WARNING_LOG, XMA_FRAME, "PrivateFrameInfo::~PrivateFrameInfo() VpiFrm not empty!");
  }
  if (m_parent) {
    m_parent->DecRef();
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateFrameInfo::~PrivateFrameInfo() - returning");
}

bool PrivateFrameInfo::IsValid() const {
  return (m_dummy == FRAME_DUMMY_VALUE);
}

std::recursive_mutex& PrivateFrameInfo::GetMutex() {
  return m_mutex;
}

VpiBufferPool* PrivateFrameInfo::GetBufferPool() {
  if (m_buffer) {
    return static_cast<VpiBufferPool*>(m_buffer->pool);
  } else if (m_pic) {
    VpiBuffer* buf = static_cast<VpiBuffer*>(m_pic->vpi_picinfo_buf);
    if (buf) {
      return static_cast<VpiBufferPool*>(buf->pool);
    }
  }
  if (m_parent) {
    IntFramePtr parent = std::dynamic_pointer_cast<InternalFrame>(m_parent);
    if (parent) {
      return parent->GetBufferPool();
    }
  }
  return nullptr;
}

void PrivateFrameInfo::SetSelf(IntFramePtr self) {
  m_self = self;
}

IntFramePtr PrivateFrameInfo::GetSelf() const {
  return m_self;
}

void PrivateFrameInfo::SetParent(FramePtr parent) {
  m_parent = parent;
  m_parent->IncRef();
}

int32_t PrivateFrameInfo::AddVpiFrm(VpiFrm* pic) {
  IntSessionFactoryPtr sf = m_sf;
  m_pic                   = vpi_frm_t(pic, [sf](VpiFrm* frm) {
    vpi_frame_unref(frm);
    if (vpi_buffer_get_ref_count(static_cast<VpiBuffer*>(frm->vpi_picinfo_buf)) == 0) {
      vpi_frame_free(sf->GetMemHandle(), frm);
    }
  });
  return XMA_SUCCESS;
}

VpiFrm* PrivateFrameInfo::GetVpiFrm() const {
  return m_pic.get();
}

int32_t PrivateFrameInfo::RemoveVpiFrm() {
  int32_t ret = m_pic ? XMA_SUCCESS : XMA_ERROR;
  m_pic.reset();
  return ret;
}

void PrivateFrameInfo::SetCallback(std::function<void(uint8_t**)> free_callback) {
  m_free_callback = free_callback;
}

void PrivateFrameInfo::CallCallback(uint8_t** frame_data) {
  if (m_free_callback) {
    m_free_callback(frame_data);
  }
}

IntSessionFactoryPtr PrivateFrameInfo::GetSessionFactory() {
  return m_sf;
}

LogPtr PrivateFrameInfo::GetLog() {
  return m_log;
}

//////////////////////////////////////////////////////////////////////////////
PrivateVpiFrmInfo::PrivateVpiFrmInfo(IntSessionFactoryPtr sf, LogPtr log) : m_sf(sf), m_log(log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::PrivateVpiFrmInfo()");
}

PrivateVpiFrmInfo::~PrivateVpiFrmInfo() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::~PrivateVpiFrmInfo()");
  if (!m_frames.empty()) {
    m_log->LogMsg(XMA_WARNING_LOG, XMA_FRAME, "PrivateVpiFrmInfo::~PrivateVpiFrmInfo() Frame list not empty!");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::~PrivateVpiFrmInfo() - returning");
}

bool PrivateVpiFrmInfo::IsValid() const {
  return (m_dummy == VPIFRM_DUMMY_VALUE);
}

int32_t PrivateVpiFrmInfo::AddFrame(FramePtr frame, VpiFrm* pic, int32_t ch) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::AddFrame(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(frame->GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(frame->GetProps()->GetSwFormat()),
      frame->GetProps()->GetWidth(), frame->GetProps()->GetHeight());
  vpi_frame_info_t vpi_info = std::make_unique<VpiFrmInfo>();
  vpi_info->frame           = frame;
  vpi_info->buf             = static_cast<VpiBuffer*>(pic->vpi_picinfo_buf);
  int32_t    planes         = frame->GetProps()->GetNumPlanes();
  int32_t    plane;
  VpiBuffer* buf;
  switch (frame->GetBufferType()) {
  case XMA_HOST_BUFFER_TYPE:
  case XMA_DEVICE_ONLY_BUFFER_TYPE:
  case XMA_DEVICE_BUFFER_TYPE: {
    buf                    = reinterpret_cast<VpiBuffer*>(pic->vpi_picinfo_buf);
    FramePtr    iter_frame = frame;
    IntFramePtr int_frame  = std::dynamic_pointer_cast<InternalFrame>(frame);
    if (int_frame) {
      XmaFrame* xma_frame = int_frame->GetFrame();
      if (xma_frame) {
        PrivateFrameInfo* frame_info = static_cast<PrivateFrameInfo*>(xma_frame->private_do_not_touch);
        if (frame_info && frame_info->IsValid()) {
          VpiFrm* frm = frame_info->GetVpiFrm();
          if (frm) {
            int32_t frm_ch;
            for (frm_ch = 0; frm_ch < VPE_CHANNEL_NUM; ++frm_ch) {
              if ((frm->ch_info[frm_ch].enabled) && (frm->ch_info[frm_ch].format == pic->ch_info[ch].format) &&
                  (frm->ch_info[frm_ch].sw_format == pic->ch_info[ch].sw_format) && (frm->ch_info[frm_ch].width == pic->ch_info[ch].width) &&
                  (frm->ch_info[frm_ch].height == pic->ch_info[ch].height)) {
                break;
              }
            }
            if (frm_ch < VPE_CHANNEL_NUM) {
              VpiBuffer* frm_buf     = static_cast<VpiBuffer*>(frm->vpi_picinfo_buf);
              buf->m_handle          = frm_buf->m_handle;
              buf->m_metadata_handle = frm_buf->m_metadata_handle;
              memcpy(buf->buf_paddr[ch], frm_buf->buf_paddr[frm_ch], sizeof(void*) * VPE_MAX_PLANES);
              memcpy(buf->buf_vaddr[ch], frm_buf->buf_vaddr[frm_ch], sizeof(void*) * VPE_MAX_PLANES);
              memcpy(buf->md_paddr[ch], frm_buf->md_paddr[frm_ch], sizeof(void*) * VPE_MAX_PLANES);
              memcpy(buf->md_vaddr[ch], frm_buf->md_vaddr[frm_ch], sizeof(void*) * VPE_MAX_PLANES);
              buf->buf_size     = frm_buf->buf_size;
              buf->md_data_size = frm_buf->md_data_size;
              m_frames.push_back(std::move(vpi_info));
              m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::AddFrame() - Returning success");
              return XMA_SUCCESS;
            }
          }
        }
      }
    }
    buf->m_handle = static_cast<VpiLinearMem*>(frame->GetBuffer(0));
    for (plane = 0; plane < planes; plane++) {
      VpiLinearMem* mem         = static_cast<VpiLinearMem*>(frame->GetBuffer(plane));
      buf->buf_paddr[ch][plane] = reinterpret_cast<void*>(mem->bus_address_ep);
      buf->buf_vaddr[ch][plane] = mem->virtual_address_rc;
    }
    break;
  }
  case NO_BUFFER:
    break;
  }
  m_frames.push_back(std::move(vpi_info));
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::AddFrame() - Returning success");
  return XMA_SUCCESS;
}

vpi_frame_info_t PrivateVpiFrmInfo::GetFrame(XmaFormatType format, XmaFormatType sw_format, int32_t width, int32_t height) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::GetFrame(format=%s, sw_format=%s, size=%dx%d)", InternalFrameProperties::GetPixFmtName(format),
      InternalFrameProperties::GetPixFmtName(sw_format), width, height);
  auto itr = m_frames.begin();
  while (itr != m_frames.end()) {
    FramePropertiesPtr props = (*itr)->frame->GetProps();
    if ((props->GetFormat() == format) && (props->GetSwFormat() == sw_format) && (props->GetWidth() == width) && (props->GetHeight() == height)) {
      vpi_frame_info_t vpi_info = std::move(*itr);
      m_frames.erase(itr);
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::GetFrame() - Returning frame");
      return vpi_info;
    }
    ++itr;
  }
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "PrivateVpiFrmInfo::GetFrame() - Returning frame not found");
  return nullptr;
}

int32_t PrivateVpiFrmInfo::GetFrameCount() const {
  return m_frames.size();
}

int32_t PrivateVpiFrmInfo::SetFreeBuf(VpePicFree func_free) {
  m_func_free = func_free;
  return XMA_SUCCESS;
}

VpePicFree PrivateVpiFrmInfo::GetFreeBuf() const {
  return m_func_free;
}

int32_t PrivateVpiFrmInfo::SetFreeData(void* data) {
  m_free_data = data;
  return XMA_SUCCESS;
}

void* PrivateVpiFrmInfo::GetFreeData() const {
  return m_free_data;
}

int32_t PrivateVpiFrmInfo::SetPool(VpiBufferPool* pool) {
  m_pool = pool;
  return XMA_SUCCESS;
}

VpiBufferPool* PrivateVpiFrmInfo::GetPool() const {
  return m_pool;
}

IntSessionFactoryPtr PrivateVpiFrmInfo::GetSessionFactory() {
  return m_sf;
}

LogPtr PrivateVpiFrmInfo::GetLog() {
  return m_log;
}

void InitVpiFrm(VpiFrm* pic) {
  pic->valid     = 0;
  pic->width     = 0;
  pic->height    = 0;
  pic->poc       = 0;
  pic->format    = VPI_FORMAT_NONE;
  pic->key_frame = 0;
  int32_t ch;
  for (ch = 0; ch < VPE_CHANNEL_NUM; ch++) {
    pic->vpi_sideband[ch] = nullptr;
  }
  memset(pic->ch_info, 0, sizeof(pic->ch_info));
  pic->pts       = 0;
  pic->dts       = 0;
  pic->next      = nullptr;
  pic->datatype  = static_cast<VpiFrmDataType>(-1);
  pic->opaque    = nullptr;
  pic->ch_index  = 0;
  pic->hdr_param = nullptr;
}

//////////////////////////////////////////////////////////////////////////////
void xma_pic_consume(void*, void* data) {
  VpiFrm* pic = static_cast<VpiFrm*>(data);

  if (!pic) {
    throw InvalidArgumentException("xma_pic_consume() Invalid parameter!");
  }

  PrivateVpiFrmInfo* info = static_cast<PrivateVpiFrmInfo*>(pic->opaque);

  if (!info || !info->IsValid()) {
    throw InvalidArgumentException("xma_pic_consume() Invalid parameter!");
  }

  LogPtr log = info->GetLog();
  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "xma_pic_consume()");

  if (pic->hdr_param) {
    for (uint8_t idx = 0; idx < pic->hdr_param->t35_param.valid_counter; ++idx) {
      delete[] pic->hdr_param->t35_param.payload_byte_data[idx];
    }
    delete pic->hdr_param;
  }

  VpiBuffer*     buffer  = static_cast<VpiBuffer*>(pic->vpi_picinfo_buf);
  VpiBufferPool* vpiPool = static_cast<VpiBufferPool*>(buffer->pool);
  BufferPool*    pool    = static_cast<BufferPool*>(vpiPool->opaque);
  if (pool) {
    pool->FreeBuffer(buffer);
  }
  vpi_memory_free_frame(info->GetSessionFactory()->GetMemHandle(), pic);
  pic->vpi_picinfo_buf = nullptr;

  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "xma_pic_consume() - Returning");
}

//////////////////////////////////////////////////////////////////////////////
void xma_free_reference_info(void*, void* data) {
  VpiFrm* pic = static_cast<VpiFrm*>(data);

  if (!pic) {
    throw InvalidArgumentException("xma_free_reference_info() Invalid parameter!");
  }

  PrivateVpiFrmInfo* info = static_cast<PrivateVpiFrmInfo*>(pic->opaque);

  if (!info || !info->IsValid()) {
    throw InvalidArgumentException("xma_free_reference_info() Invalid parameter!");
  }

  LogPtr log = info->GetLog();
  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "xma_free_reference_info()");
  IntSessionFactoryPtr sf = info->GetSessionFactory();

  VpiBuffer* buf = static_cast<VpiBuffer*>(pic->vpi_picinfo_buf);
  if (buf) {
    buf->pool    = info->GetPool();
    buf->freebuf = info->GetFreeBuf();
    if (buf->freebuf) {
      buf->freebuf(buf->opaque, buf->data);
    }
  }

  for (int32_t ch = 0; ch < VPE_CHANNEL_NUM; ++ch) {
    VpiChannelInfo& ch_info = pic->ch_info[ch];
    if (ch_info.enabled) {
      vpi_frame_info_t vpi_info = info->GetFrame(
          static_cast<XmaFormatType>(ch_info.format), static_cast<XmaFormatType>(ch_info.sw_format), ch_info.width, ch_info.height);
      if (vpi_info) {
        pic->vpi_picinfo_buf = vpi_info->buf;
      }
      memset(&ch_info, 0, sizeof(VpiChannelInfo));
    }
  }

  delete info;

  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "xma_free_reference_info() - Returning");
}

//////////////////////////////////////////////////////////////////////////////
InternalFrameProperties::InternalFrameProperties(
    IntSessionFactoryPtr sf, LogPtr log, enum XmaFormatType format, enum XmaFormatType sw_format, int32_t width, int32_t height, uint32_t flags) :
    m_sf(sf),
    m_log(log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrameProperties::InternalFrameProperties(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(format), InternalFrameProperties::GetPixFmtName(sw_format), width, height);
  memset(&m_props, 0, sizeof(XmaFrameProperties));
  m_props.format    = format;
  m_props.sw_format = sw_format;
  m_props.width     = width;
  m_props.height    = height;
  m_props.flags     = flags;

  int32_t planes = GetNumPlanes();
  int32_t plane;
  for (plane = 0; plane < planes; ++plane) {
    m_props.linesize[plane] = vpi_frame_get_plane_stride(m_props.sw_format, plane, m_props.width, XMA_FRAME_W_ALIGN);
  }
  for (; plane < XMA_MAX_PLANES; ++plane) {
    m_props.linesize[plane] = 0;
  }
}

///////////////////////////////////////////////////////////////////////////////
InternalFrameProperties::InternalFrameProperties(IntSessionFactoryPtr sf, LogPtr log, XmaFrameProperties* props) : m_props(*props), m_sf(sf), m_log(log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrameProperties::InternalFrameProperties(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(props->format), InternalFrameProperties::GetPixFmtName(props->sw_format), props->width, props->height);
  int32_t planes = GetNumPlanes();
  int32_t plane;
  for (plane = 0; plane < planes; ++plane) {
    if (m_props.linesize[plane] == 0) {
      m_props.linesize[plane] = vpi_frame_get_plane_stride(m_props.sw_format, plane, m_props.width, XMA_FRAME_W_ALIGN);
    }
  }
  for (; plane < XMA_MAX_PLANES; ++plane) {
    m_props.linesize[plane] = 0;
  }
}

///////////////////////////////////////////////////////////////////////////////
XmaFrameProperties InternalFrameProperties::GetFrameProperties() const {
  return m_props;
}

///////////////////////////////////////////////////////////////////////////////
char* InternalFrameProperties::GetPixFmtName(XmaFormatType fmt) {
  return const_cast<char*>(pixel_format_names[fmt]);
}

///////////////////////////////////////////////////////////////////////////////
enum XmaFormatType InternalFrameProperties::GetFormat() const {
  return m_props.format;
}

///////////////////////////////////////////////////////////////////////////////
enum XmaFormatType InternalFrameProperties::GetSwFormat() const {
  return m_props.sw_format;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrameProperties::GetWidth() const {
  return m_props.width;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrameProperties::GetHeight() const {
  return m_props.height;
}

///////////////////////////////////////////////////////////////////////////////
uint32_t InternalFrameProperties::GetFlags() const {
  return m_props.flags;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrameProperties::GetNumPlanes() const {
  return vpi_frame_get_planes(m_props.sw_format);
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrameProperties::GetPlaneHeight(int32_t plane) const {
  if ((plane < 0) || (plane > GetNumPlanes())) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrameProperties::GetPlaneHeight() Invalid parameter!");
  }
  return vpi_frame_get_plane_height(m_props.sw_format, plane, m_props.height, XMA_FRAME_H_ALIGN);
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrameProperties::GetPlaneStride(int32_t plane) const {
  if ((plane < 0) || (plane > GetNumPlanes())) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrameProperties::GetPlaneStride() Invalid parameter!");
  }
  return m_props.linesize[plane];
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrameProperties::GetPlaneSize(int32_t plane) const {
  if ((plane < 0) || (plane > GetNumPlanes())) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrameProperties::GetPlaneSize() Invalid parameter!");
  }
  return m_props.linesize[plane] * GetPlaneHeight(plane);
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrameProperties::GetBufferSize(const FramePropertiesPtr props) {
  int32_t planes = props->GetNumPlanes();
  int32_t size   = 1024; // Needed for AV1?
  for (int32_t plane = 0; plane < planes; ++plane) {
    size += props->GetPlaneSize(plane);
  }
  return size;
}

///////////////////////////////////////////////////////////////////////////////
InternalFrame::InternalFrame(
    IntSessionFactoryPtr sf, LogPtr log, FramePropertiesPtr props, bool dummy, buffer_t buffer, XmaFraction frame_rate, int64_t pts, bool is_idr) :
    m_props(props),
    m_sf(sf), m_log(log) {
  if (dummy) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame(dummy)");
  } else {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame(format=%s, sw_format=%s, size=%dx%d)",
        InternalFrameProperties::GetPixFmtName(props->GetFormat()), InternalFrameProperties::GetPixFmtName(props->GetSwFormat()), props->GetWidth(), props->GetHeight());
    if (!buffer) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::InternalFrame() No pool provided!");
    }
  }
  m_frame = std::make_unique<XmaFrame>();

  memset(m_frame.get(), 0, sizeof(XmaFrame));

  m_frame->frame_props.format    = props->GetFormat();
  m_frame->frame_props.sw_format = props->GetSwFormat();
  m_frame->frame_props.width     = props->GetWidth();
  m_frame->frame_props.height    = props->GetHeight();
  m_frame->frame_props.flags     = props->GetFlags();
  m_frame->frame_rate            = frame_rate;
  m_frame->pts                   = pts;
  m_frame->is_idr                = is_idr ? 1 : 0;

  int32_t planes = props->GetNumPlanes();
  int32_t plane;
  if (dummy) {
    for (plane = 0; plane < planes; ++plane) {
      m_frame->frame_props.linesize[plane] = props->GetPlaneStride(plane);
      m_frame->data[plane].buffer_type     = NO_BUFFER;
      m_frame->data[plane].buffer          = nullptr;
      m_frame->data[plane].host            = nullptr;
      m_frame->data[plane].device          = nullptr;
    }
    for (; plane < XMA_MAX_PLANES; ++plane) {
      m_frame->frame_props.linesize[plane] = 0;
      m_frame->data[plane].buffer_type     = NO_BUFFER;
      m_frame->data[plane].buffer          = nullptr;
      m_frame->data[plane].host            = nullptr;
      m_frame->data[plane].device          = nullptr;
    }
  } else {
    int32_t       sizes[XMA_MAX_PLANES];
    VpiLinearMem* mem;
    VpiLinearMem* mem0;
    for (plane = 0; plane < planes; ++plane) {
      sizes[plane]                         = props->GetPlaneSize(plane);
      m_frame->frame_props.linesize[plane] = props->GetPlaneStride(plane);
      m_frame->data[plane].buffer_type     = (props->GetFormat() == XMA_VPE_FMT_TYPE) ? XMA_DEVICE_ONLY_BUFFER_TYPE : XMA_HOST_BUFFER_TYPE;
      mem                                  = static_cast<VpiLinearMem*>(vpi_memory_calloc(1, sizeof(VpiLinearMem)));
      if (!mem) {
        for (--plane; plane >= 0; --plane) {
          vpi_memory_free(m_frame->data[plane].buffer);
        }
        THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::InternalFrame() Out Of Memory!");
      }
      m_frame->data[plane].buffer   = mem;
      m_frame->data[plane].is_clone = false;
    }
    for (; plane < XMA_MAX_PLANES; ++plane) {
      m_frame->frame_props.linesize[plane] = 0;
      m_frame->data[plane].buffer_type     = NO_BUFFER;
      m_frame->data[plane].buffer          = nullptr;
      m_frame->data[plane].host            = nullptr;
      m_frame->data[plane].device          = nullptr;
    }
    mem0                    = buffer->m_handle;
    mem                     = static_cast<VpiLinearMem*>(m_frame->data[0].buffer);
    mem->size               = sizes[0]; // cppcheck-suppress [uninitvar]
    mem->virtual_address_rc = mem0->virtual_address_rc;
    mem->bus_address_ep     = mem0->bus_address_ep;
    mem->task_id            = mem0->task_id;
    mem->mem_type           = mem0->mem_type;
    m_frame->data[0].host   = static_cast<uint8_t*>(mem->virtual_address_rc);
    m_frame->data[0].device = reinterpret_cast<void*>(mem->bus_address_ep);
    for (plane = 1; plane < planes; ++plane) {
      mem0                        = static_cast<VpiLinearMem*>(m_frame->data[plane - 1].buffer);
      mem                         = static_cast<VpiLinearMem*>(m_frame->data[plane].buffer);
      mem->size                   = sizes[plane];
      mem->bus_address_ep         = mem0->bus_address_ep + sizes[plane - 1];
      mem->virtual_address_rc     = static_cast<uint8_t*>(mem0->virtual_address_rc) + sizes[plane - 1];
      mem->task_id                = mem0->task_id;
      mem->mem_type               = mem0->mem_type;
      m_frame->data[plane].host   = static_cast<uint8_t*>(mem->virtual_address_rc);
      m_frame->data[plane].device = reinterpret_cast<void*>(mem->bus_address_ep);
    }
  }

  auto info                     = std::make_unique<PrivateFrameInfo>(sf, m_log, std::move(buffer));
  m_frame->private_do_not_touch = static_cast<void*>(info.release());

  IncRef();

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_frames.push_back(m_frame.get());
  lock.unlock();
#endif
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalFrame::InternalFrame(IntSessionFactoryPtr sf, LogPtr log, FramePropertiesPtr props, uint8_t** frame_data, std::function<void(uint8_t**)> free_callback,
    XmaFraction frame_rate, int64_t pts, bool is_idr) :
    m_props(props),
    m_sf(sf), m_log(log) {
  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame(format=%s, sw_format=%s, size=%dx%d)", InternalFrameProperties::GetPixFmtName(props->GetFormat()),
      InternalFrameProperties::GetPixFmtName(props->GetSwFormat()), props->GetWidth(), props->GetHeight());
  m_frame = std::make_unique<XmaFrame>();
  memset(m_frame.get(), 0, sizeof(XmaFrame));
  m_frame->frame_props.format    = props->GetFormat();
  m_frame->frame_props.sw_format = props->GetSwFormat();
  m_frame->frame_props.width     = props->GetWidth();
  m_frame->frame_props.height    = props->GetHeight();
  m_frame->frame_props.flags     = props->GetFlags();
  m_frame->frame_rate            = frame_rate;
  m_frame->pts                   = pts;
  m_frame->is_idr                = is_idr;

  auto info = std::make_unique<PrivateFrameInfo>(sf, log);
  info->SetCallback(free_callback);

  int32_t planes = props->GetNumPlanes();
  int32_t plane;
  for (plane = 0; plane < planes; ++plane) {
    m_frame->frame_props.linesize[plane] = props->GetPlaneStride(plane);
    m_frame->data[plane].buffer_type     = XMA_HOST_BUFFER_TYPE;
    m_frame->data[plane].buffer          = nullptr;
    m_frame->data[plane].host            = frame_data[plane];
    m_frame->data[plane].device          = nullptr;
    m_frame->data[plane].is_clone        = true;
  }
  for (; plane < XMA_MAX_PLANES; ++plane) {
    m_frame->frame_props.linesize[plane] = 0;
    m_frame->data[plane].buffer_type     = NO_BUFFER;
    m_frame->data[plane].buffer          = nullptr;
    m_frame->data[plane].host            = nullptr;
    m_frame->data[plane].device          = nullptr;
  }

  m_frame->private_do_not_touch = static_cast<void*>(info.release());

  IncRef();

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_frames.push_back(m_frame.get());
  lock.unlock();
#endif
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalFrame::InternalFrame(IntSessionFactoryPtr sf, LogPtr log, FramePtr frame, XmaFraction frame_rate, int64_t pts, bool is_idr) :
    m_props(frame->GetProps()), m_sf(sf), m_log(log) {
  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame(format=%s, sw_format=%s, size=%dx%d)", InternalFrameProperties::GetPixFmtName(m_props->GetFormat()),
      InternalFrameProperties::GetPixFmtName(m_props->GetSwFormat()), m_props->GetWidth(), m_props->GetHeight());
  m_frame = std::make_unique<XmaFrame>();
  memset(m_frame.get(), 0, sizeof(XmaFrame));
  m_frame->frame_props.format    = m_props->GetFormat();
  m_frame->frame_props.sw_format = m_props->GetSwFormat();
  m_frame->frame_props.width     = m_props->GetWidth();
  m_frame->frame_props.height    = m_props->GetHeight();
  m_frame->frame_props.flags     = m_props->GetFlags();
  m_frame->frame_rate            = frame_rate;
  m_frame->pts                   = pts;
  m_frame->is_idr                = is_idr;

  auto info = std::make_unique<PrivateFrameInfo>(sf, log);
  info->SetParent(frame);

  int32_t planes = m_props->GetNumPlanes();
  int32_t plane;
  for (plane = 0; plane < planes; ++plane) {
    m_frame->frame_props.linesize[plane] = m_props->GetPlaneStride(plane);
    m_frame->data[plane].buffer_type     = frame->GetBufferType();
    m_frame->data[plane].buffer          = frame->GetBuffer(plane);
    m_frame->data[plane].host            = frame->GetHostBuffer(plane);
    m_frame->data[plane].device          = frame->GetDeviceBuffer(plane);
    m_frame->data[plane].is_clone        = true;
  }
  for (; plane < XMA_MAX_PLANES; ++plane) {
    m_frame->frame_props.linesize[plane] = 0;
    m_frame->data[plane].buffer_type     = NO_BUFFER;
    m_frame->data[plane].buffer          = nullptr;
    m_frame->data[plane].host            = nullptr;
    m_frame->data[plane].device          = nullptr;
  }

  m_frame->private_do_not_touch = static_cast<void*>(info.release());

  IncRef();

  for (int32_t type = XMA_FRAME_SIDE_DATA_START; type < XMA_FRAME_SIDE_DATA_MAX_COUNT; ++type) {
    SideDataPtr sd;
    if ((sd = frame->GetSideData(static_cast<XmaFrameSideDataType>(type))) != nullptr) {
      AddSideData(sd);
    }
  }

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_frames.push_back(m_frame.get());
  lock.unlock();
#endif
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalFrame::InternalFrame(
    IntSessionFactoryPtr sf, LogPtr log, VpiFrm* pic, XmaFraction frame_rate, XmaFormatType format, XmaFormatType sw_format, int32_t width, int32_t height) :
    m_sf(sf),
    m_log(log) {
  if (!pic->valid) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::InternalFrame() Invalid pic!");
  }

  if ((width == 0) || (height == 0)) {
    width  = pic->width;
    height = pic->height;
  }

  int32_t ch;
  for (ch = 0; ch < VPE_CHANNEL_NUM; ++ch) {
    if (pic->ch_info[ch].enabled && (pic->ch_info[ch].width == width) && (pic->ch_info[ch].height == height) && (pic->ch_info[ch].format == pic->format)) {
      break;
    }
  }
  if (ch == VPE_CHANNEL_NUM) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::InternalFrame() Unable to find channel!");
  }

  if ((format == XMA_NONE_FMT_TYPE) || (sw_format == XMA_NONE_FMT_TYPE)) {
    format    = static_cast<XmaFormatType>(pic->ch_info[ch].format);
    sw_format = static_cast<XmaFormatType>(pic->ch_info[ch].sw_format);
  } else {
    for (; ch < VPE_CHANNEL_NUM; ++ch) {
      if (pic->ch_info[ch].enabled && (static_cast<XmaFormatType>(pic->ch_info[ch].format) == format) &&
          (static_cast<XmaFormatType>(pic->ch_info[ch].sw_format) == sw_format) && (pic->ch_info[ch].width == width) && (pic->ch_info[ch].height == height) &&
          (pic->ch_info[ch].format == pic->format)) {
        break;
      }
    }
    if (ch == VPE_CHANNEL_NUM) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::InternalFrame() Unable to find channel!");
    }
  }
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame(format=%s, sw_format=%s, size=%dx%d)", InternalFrameProperties::GetPixFmtName(format),
      InternalFrameProperties::GetPixFmtName(sw_format), width, height);

  int32_t planes = vpi_frame_get_planes(sw_format);

  VpiBuffer* vpi_buf = static_cast<VpiBuffer*>(pic->vpi_picinfo_buf);
  if (vpi_buf) {
    PrivateVpiFrmInfo* info = static_cast<PrivateVpiFrmInfo*>(pic->opaque);
    if (info && info->IsValid()) {
      vpi_frame_info_t vpi_info = info->GetFrame(format, sw_format, width, height);
      if (vpi_info) {
        m_frame = std::make_unique<XmaFrame>();
        memset(m_frame.get(), 0, sizeof(XmaFrame));

        m_frame->frame_props.format    = format;
        m_frame->frame_props.sw_format = sw_format;
        m_frame->frame_props.width     = width;
        m_frame->frame_props.height    = height;
        m_frame->frame_props.flags     = pic->ch_info[ch].flag;
        m_frame->frame_rate            = frame_rate;
        m_frame->pts                   = pic->pts;
        m_frame->is_idr                = pic->key_frame;

        int32_t plane;
        for (plane = 0; plane < XMA_MAX_PLANES; ++plane) {
          m_frame->frame_props.linesize[plane] = 0;
          m_frame->data[plane].buffer_type     = NO_BUFFER;
          m_frame->data[plane].buffer          = nullptr;
          m_frame->data[plane].host            = nullptr;
          m_frame->data[plane].device          = nullptr;
        }

        CopyFrame(m_log, m_frame.get(), std::move(vpi_info->frame));
        CopySideData(pic);
        m_props = std::make_shared<InternalFrameProperties>(m_sf, m_log, &m_frame->frame_props);
        if (info->GetFrameCount() == 0) {
          vpi_buf->freebuf = info->GetFreeBuf();
          pic->opaque      = nullptr;
          delete info;
        }

#ifdef XMA_CHECK_MEM
        std::unique_lock<std::mutex> lock(m_sf->m_mutex);
        m_sf->m_frames.push_back(m_frame.get());
        lock.unlock();
#endif
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame() - Returning");
        return;
      }
    }
  }

  m_frame   = std::make_unique<XmaFrame>();
  auto info = std::make_unique<PrivateFrameInfo>(sf, log);

  memset(m_frame.get(), 0, sizeof(XmaFrame));

  XmaBufferType buffer_type;
  if (vpi_buf->m_handle->virtual_address_rc) {
    if (vpi_buf->m_handle->bus_address_ep) {
      buffer_type = XMA_DEVICE_ONLY_BUFFER_TYPE;
    } else {
      buffer_type = XMA_HOST_BUFFER_TYPE;
    }
  } else {
    if (vpi_buf->m_handle->bus_address_ep) {
      buffer_type = XMA_DEVICE_ONLY_BUFFER_TYPE;
    } else {
      buffer_type = NO_BUFFER;
    }
  }

  int32_t plane;
  for (plane = 0; plane < planes; ++plane) {
    VpiLinearMem* mem                    = static_cast<VpiLinearMem*>(vpi_memory_calloc(1, sizeof(VpiLinearMem)));
    mem->size                            = vpi_frame_get_plane_size(pic->ch_info[ch].sw_format, plane, width, pic->ch_info[ch].walign, height, pic->ch_info[ch].halign);
    mem->request_size                    = mem->size;
    mem->virtual_address_rc              = vpi_buf->buf_vaddr[ch][plane];
    mem->bus_address_ep                  = reinterpret_cast<addr_t>(vpi_buf->buf_paddr[ch][plane]);
    m_frame->data[plane].buffer          = mem;
    m_frame->data[plane].buffer_type     = buffer_type;
    m_frame->data[plane].host            = static_cast<uint8_t*>(mem->virtual_address_rc);
    m_frame->data[plane].device          = reinterpret_cast<void*>(mem->bus_address_ep);
    m_frame->data[plane].is_clone        = false;
    m_frame->frame_props.linesize[plane] = pic->ch_info[ch].stride_size[plane];
  }
  for (; plane < XMA_MAX_PLANES; ++plane) {
    m_frame->frame_props.linesize[plane] = 0;
    m_frame->data[plane].buffer_type     = NO_BUFFER;
    m_frame->data[plane].buffer          = nullptr;
    m_frame->data[plane].host            = nullptr;
    m_frame->data[plane].device          = nullptr;
  }

  m_frame->frame_props.format    = static_cast<XmaFormatType>(pic->ch_info[ch].format);
  m_frame->frame_props.sw_format = static_cast<XmaFormatType>(pic->ch_info[ch].sw_format);
  m_frame->frame_props.width     = width;
  m_frame->frame_props.height    = height;
  m_frame->frame_props.flags     = pic->ch_info[ch].flag;
  m_frame->frame_rate            = frame_rate;
  m_frame->pts                   = pic->pts;
  m_frame->is_idr                = pic->key_frame;
  info->AddVpiFrm(pic);

  m_props = std::make_shared<InternalFrameProperties>(m_sf, m_log, &m_frame->frame_props);

  m_frame->private_do_not_touch = static_cast<void*>(info.release());

  IncRef();
  CopySideData(pic);

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_frames.push_back(m_frame.get());
  lock.unlock();
#endif
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::InternalFrame() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalFrame::~InternalFrame() {
  if (!m_frame) {
    return;
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::~InternalFrame(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height);

  DecRef();

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::~InternalFrame() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
XmaFrame* InternalFrame::GetFrame() {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetFrame", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetFrame(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetFrame() - Returning");

  return m_frame.get();
}

///////////////////////////////////////////////////////////////////////////////
VpiBufferPool* InternalFrame::GetBufferPool() {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetBufferPool", XMA_FRAME, m_frame);

  PrivateFrameInfo* info = static_cast<PrivateFrameInfo*>(m_frame->private_do_not_touch);
  if ((info) && (info->IsValid())) {
    return info->GetBufferPool();
  }

  return nullptr;
}

///////////////////////////////////////////////////////////////////////////////
void InternalFrame::ExtendBufferPool(uint32_t frames) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::ExtendBufferPool", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::ExtendBufferPool(format=%s, sw_format=%s, size=%dx%d, extend=%u)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height, frames);

  VpiBufferPool*    pool = nullptr;
  PrivateFrameInfo* info = static_cast<PrivateFrameInfo*>(m_frame->private_do_not_touch);
  if ((info) && (info->IsValid())) {
    pool = info->GetBufferPool();
  }

  VpiRet ret = VPI_ERR_SW;
  if (pool) {
    int32_t max_capacity = vpi_buffer_pool_get_capacity(pool);
    int32_t curr_depth   = vpi_buffer_pool_get_depth(pool);
    int32_t extend_num   = frames;
    int32_t new_depth    = curr_depth + extend_num;
    if (max_capacity) {
      if (new_depth > max_capacity) {
        extend_num = max_capacity - curr_depth;
      }
    }
    ret = vpi_buffer_pool_extend(pool, extend_num);
  }
  if (ret == VPI_SUCCESS) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::ExtendBufferPool() - returned success");
  } else {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::ExtendBufferPool() - returned");
  }
}

///////////////////////////////////////////////////////////////////////////////
void InternalFrame::SetSelf() {
  LOCK_XMA_FRAME(m_log, "InternalFrame::SetSelf", XMA_FRAME, m_frame);

  PrivateFrameInfo* info = static_cast<PrivateFrameInfo*>(m_frame->private_do_not_touch);
  if ((info) && (info->IsValid())) {
    info->SetSelf(std::dynamic_pointer_cast<InternalFrame>(shared_from_this()));
  }
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrame::IncRef() {
  LOCK_XMA_FRAME(m_log, "InternalFrame::IncRef", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::IncRef(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height);

  int32_t planes = m_props->GetNumPlanes();
  for (int32_t plane = 0; plane < planes; ++plane) {
    ++m_frame->data[plane].refcount;
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::IncRef() - Returning");
  return m_frame->data[0].refcount;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrame::DecRef() {
  LOCK_XMA_FRAME(m_log, "InternalFrame::DecRef", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::DecRef(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height);

  int32_t planes = m_props->GetNumPlanes();
  for (int32_t plane = 0; plane < planes; ++plane) {
    --m_frame->data[plane].refcount;
  }

  if (m_frame->data[0].refcount > 0) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::DecRef() - Returning");
    return m_frame->data[0].refcount;
  }

  if (m_frame->private_do_not_touch) {
    PrivateFrameInfo* info = static_cast<PrivateFrameInfo*>(m_frame->private_do_not_touch);
    if (info->IsValid()) {
      ClearAllSideData();
      if (info->RemoveVpiFrm() == XMA_SUCCESS) { // if the frame was made from a VpiFrm...

        m_frame->private_do_not_touch = nullptr;
        for (int32_t plane = 0; plane < planes; ++plane) {
          vpi_memory_free(m_frame->data[plane].buffer);
        }
        memset(m_frame->data, 0, sizeof(XmaBufferRef) * XMA_MAX_PLANES);

#ifdef XMA_CHECK_MEM
        std::unique_lock<std::mutex> lock(m_sf->m_mutex);
        m_sf->m_frames.remove(m_frame.get());
        lock.unlock();
#endif

        guard.unlock();
        delete info;
        m_frame = nullptr;

        m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::DecRef() - Returning");
        return 0;
      }
      if (m_frame->data[0].is_clone) {
        uint8_t* data[XMA_MAX_PLANES];
        memset(&data, 0, sizeof(data));
        for (int32_t plane = 0; plane < planes; ++plane) {
          data[plane] = m_frame->data[plane].host;
        }
        info->CallCallback(data);
      }
      guard.unlock();
      delete info;
      m_frame->private_do_not_touch = nullptr;
    }
  }

  switch (m_frame->data[0].buffer_type) {
  case XMA_HOST_BUFFER_TYPE:
  case XMA_DEVICE_ONLY_BUFFER_TYPE:
  case XMA_DEVICE_BUFFER_TYPE:
    for (int32_t plane = 0; plane < planes; ++plane) {
      if ((m_frame->data[plane].buffer) && (!m_frame->data[plane].is_clone)) {
        vpi_memory_free(static_cast<void*>(m_frame->data[plane].buffer));
      }

      m_frame->data[plane].buffer      = nullptr;
      m_frame->data[plane].host        = nullptr;
      m_frame->data[plane].device      = nullptr;
      m_frame->data[plane].buffer_type = NO_BUFFER;
    }
    break;
  case NO_BUFFER:
    break;
  default:
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::DecRef() Invalid buffer type!");
  }

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_frames.remove(m_frame.get());
  lock.unlock();
#endif

  m_frame = nullptr;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::DecRef() - Returning");
  return 0;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrame::GetRefCount() {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetRefCount", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetRefCount(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(m_frame->frame_props.format), InternalFrameProperties::GetPixFmtName(m_frame->frame_props.sw_format),
      m_frame->frame_props.width, m_frame->frame_props.height);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetRefCount() - Returning");
  return m_frame->data[0].refcount;
}

///////////////////////////////////////////////////////////////////////////////
XmaBufferType InternalFrame::GetBufferType() const {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetBufferType", XMA_FRAME, m_frame);

  return m_frame->data[0].buffer_type;
}

///////////////////////////////////////////////////////////////////////////////
void* InternalFrame::GetBuffer(int32_t plane) const {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetBuffer", XMA_FRAME, m_frame);

  if ((plane < 0) || (plane >= m_props->GetNumPlanes())) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetBuffer() Invalid parameter!");
  }

  return m_frame->data[plane].buffer;
}

///////////////////////////////////////////////////////////////////////////////
uint8_t* InternalFrame::GetHostBuffer(int32_t plane) const {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetHostBuffer", XMA_FRAME, m_frame);

  if ((plane < 0) || (plane >= m_props->GetNumPlanes())) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetHostBuffer() Invalid parameter!");
  }

  switch (m_frame->data[0].buffer_type) {
  case NO_BUFFER:
  case XMA_DEVICE_ONLY_BUFFER_TYPE:
    return nullptr;
  case XMA_HOST_BUFFER_TYPE:
  case XMA_DEVICE_BUFFER_TYPE:
    return m_frame->data[plane].host;
  default:
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetHostBuffer() Invalid buffer type!");
  }
}

///////////////////////////////////////////////////////////////////////////////
void* InternalFrame::GetDeviceBuffer(int32_t plane) const {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetDeviceBuffer", XMA_FRAME, m_frame);

  if ((plane < 0) || (plane >= m_props->GetNumPlanes())) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetDeviceBuffer() Invalid parameter!");
  }

  switch (m_frame->data[0].buffer_type) {
  case NO_BUFFER:
  case XMA_HOST_BUFFER_TYPE:
    return nullptr;
  case XMA_DEVICE_ONLY_BUFFER_TYPE:
  case XMA_DEVICE_BUFFER_TYPE:
    return m_frame->data[plane].device;
  default:
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::GetDeviceBuffer() Invalid buffer type!");
  }
}

///////////////////////////////////////////////////////////////////////////////
FramePropertiesPtr InternalFrame::GetProps() const {
  return m_props;
}

///////////////////////////////////////////////////////////////////////////////
XmaFraction InternalFrame::GetFrameRate() const {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetFrameRate", XMA_FRAME, m_frame);

  return m_frame->frame_rate;
}

///////////////////////////////////////////////////////////////////////////////
int64_t InternalFrame::GetPts() const {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetPts", XMA_FRAME, m_frame);

  return m_frame->pts;
}

///////////////////////////////////////////////////////////////////////////////
int InternalFrame::IsIdr() const {
  LOCK_XMA_FRAME(m_log, "InternalFrame::IsIdr", XMA_FRAME, m_frame);

  return m_frame->is_idr;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrame::AddSideData(SideDataPtr side_data) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::AddSideData", XMA_FRAME, m_frame);

  if (!side_data) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::AddSideData() Invalid parameter!");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::AddSideData(frame=(format=%s, sw_format=%s, size=%dx%d), side data=(type=%d))",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight(), static_cast<int>(side_data->GetType()));

  if (!m_frame->side_data) {
    m_frame->side_data = new XmaFrameSideData*[XMA_FRAME_SIDE_DATA_MAX_COUNT];
    if (!m_frame->side_data) {
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::AddSideData() Out of memory!");
    }
    memset(m_frame->side_data, 0, sizeof(XmaFrameSideData*) * XMA_FRAME_SIDE_DATA_MAX_COUNT);
  }

  RemoveSideDataType(side_data->GetType());

  IntSideDataPtr sd = std::dynamic_pointer_cast<InternalSideData>(side_data);
  if (!sd) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::AddSideData() side data was not created by SessionFactory::AllocSideData()!");
  }
  m_frame->side_data[side_data->GetType()] = sd->GetSideData();
  sd->IncRef();

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::AddSideData() - Returning");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
SideDataPtr InternalFrame::GetSideData(enum XmaFrameSideDataType type) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::GetSideData", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetSideData(frame=(format=%s, sw_format=%s, size=%dx%d), side data=(type=%d))",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight(), static_cast<int>(type));

  if (!m_frame->side_data) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetSideData() - Returning side data not found");
    return nullptr;
  }

  XmaFrameSideData* side_data = m_frame->side_data[type];
  if (!side_data) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetSideData() - Returning side data not found");
    return nullptr;
  }

  PrivateSideDataInfo* sd_info = static_cast<PrivateSideDataInfo*>(side_data->private_do_not_touch);
  IntSideDataPtr       ret;
  if ((sd_info) && (sd_info->IsValid())) {
    ret = sd_info->GetSelf();
    if (!ret) {
      InternalSideData* sd = sd_info->GetSelfPtr();
      if (sd) {
        ret = std::dynamic_pointer_cast<InternalSideData>(sd->shared_from_this());
      }
    }
  }
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::GetSideData() - Returning side data");
  return ret;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrame::RemoveSideData(SideDataPtr side_data) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::RemoveSideData", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::RemoveSideData()");

  if (!side_data) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::RemoveSideData() Invalid parameter!");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::RemoveSideData(frame=(format=%s, sw_format=%s, size=%dx%d), side data=(type=%d))",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight(), static_cast<int>(side_data->GetType()));

  enum XmaFrameSideDataType type          = side_data->GetType();
  SideDataPtr               old_side_data = GetSideData(type);
  if (!old_side_data) {
    THROW_INVALID_ARGUMENT_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::RemoveSideData() Frame %p has no side data buffer %p!", m_frame.get(), side_data.get());
  }
  if (side_data->GetBufferType() != old_side_data->GetBufferType()) {
    THROW_INVALID_ARGUMENT_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::RemoveSideData() new side data buffer type does not match old side data buffer type!");
  }
  switch (side_data->GetBufferType()) {
  case XMA_HOST_BUFFER_TYPE:
    if (side_data->GetHostPtr() != old_side_data->GetHostPtr()) {
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::RemoveSideData() new side data host buffer pointer does not match old side data host buffer pointer!");
    }
    break;
  case XMA_DEVICE_BUFFER_TYPE:
    if ((side_data->GetHostPtr() != old_side_data->GetHostPtr()) || (side_data->GetDevicePtr() != old_side_data->GetDevicePtr())) {
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::RemoveSideData() new side data buffer pointers do not match old side data buffer pointers!");
    }
    break;
  case XMA_DEVICE_ONLY_BUFFER_TYPE:
    if (side_data->GetDevicePtr() != old_side_data->GetDevicePtr()) {
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::RemoveSideData() new side data device buffer pointer does not match old side data device buffer pointer!");
    }
    break;
  case NO_BUFFER:
    break;
  }

  RemoveSideDataType(type);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::RemoveSideData() - Returning");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalFrame::RemoveSideDataType(enum XmaFrameSideDataType type) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::RemoveSideDataType", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::RemoveSideDataType(frame=(format=%s, sw_format=%s, size=%dx%d), type=%d)",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight(), static_cast<int>(type));

  if (!m_frame->side_data) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::RemoveSideDataType() Frame %p has no side data of type %d!", m_frame.get(), type);
  }

  XmaFrameSideData* side_data = m_frame->side_data[type];
  if (side_data) {
    PrivateSideDataInfo* sd_info = static_cast<PrivateSideDataInfo*>(side_data->private_do_not_touch);
    if ((sd_info) && (sd_info->IsValid())) {
      IntSideDataPtr sd;
      sd = sd_info->GetSelf();
      if (sd) {
        sd->DecRef();
      } else {
        InternalSideData* sd_ptr = sd_info->GetSelfPtr();
        if (sd_ptr) {
          sd_ptr->DecRef();
        }
      }
    }
    m_frame->side_data[type] = nullptr;
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::RemoveSideDataType() - Returning");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
void InternalFrame::ClearAllSideData() {
  LOCK_XMA_FRAME(m_log, "InternalFrame::ClearAllSideData", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::ClearAllSideData(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight());

  if (m_frame->side_data) {
    for (int32_t i = 0; i < XMA_FRAME_SIDE_DATA_MAX_COUNT; ++i) {
      if (m_frame->side_data[i]) {
        RemoveSideDataType(static_cast<XmaFrameSideDataType>(i));
      }
    }
    delete[] m_frame->side_data;
    m_frame->side_data = nullptr;
  }
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::ClearAllSideData() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
void InternalFrame::CopySideData(VpiFrm* src) {
  LOCK_XMA_FRAME(m_log, "InternalFrame::CopySideData", XMA_FRAME, m_frame);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "InternalFrame::CopySideData(format=%s, sw_format=%s, size=%dx%d)",
      InternalFrameProperties::GetPixFmtName(GetProps()->GetFormat()), InternalFrameProperties::GetPixFmtName(GetProps()->GetSwFormat()), GetProps()->GetWidth(),
      GetProps()->GetHeight());

  // always add XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS
  IntSideDataPtr sd = std::make_shared<InternalSideData>(m_sf, m_log, XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS, XMA_HOST_BUFFER_TYPE, sizeof(XmaDynamicEncParams));
  if (!sd) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideData() Out of memory!");
  }
  memcpy(sd->GetHostPtr(), &src->dynamic_params, sizeof(XmaDynamicEncParams));
  AddSideData(sd);
  sd->DecRef();
  XmaFrameSideData*    side_data = m_frame->side_data[XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS];
  PrivateSideDataInfo* sd_info   = static_cast<PrivateSideDataInfo*>(side_data->private_do_not_touch);
  if ((sd_info) && (sd_info->IsValid())) {
    sd_info->SetSelf(sd);
  }

  // add XMA_FRAME_SIDE_DATA_HDR10_PARAMS
  if (src->hdr_param) {
    if (sizeof(VpiHdrParameters) != sizeof(XmaHdrParams)) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideData() HDR sidedata is incorrect size, must be same size as VSI HDR sidedata!");
    }
    sd = std::make_shared<InternalSideData>(m_sf, m_log, XMA_FRAME_SIDE_DATA_HDR10_PARAMS, XMA_HOST_BUFFER_TYPE, sizeof(XmaHdrParams));
    if (!sd) {
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideData() Out of memory!");
    }
    XmaHdrParams* params = reinterpret_cast<XmaHdrParams*>(sd->GetHostPtr());
    memcpy(params, src->hdr_param, sizeof(XmaHdrParams));
    for (uint8_t idx = 0; idx < src->hdr_param->t35_param.valid_counter; ++idx) {
      params->t35.payload_byte_data[idx] = new uint8_t[src->hdr_param->t35_param.payload_byte_length[idx]];
      if (!params->t35.payload_byte_data[idx]) {
        THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideData() Out of memory!");
      }
      memcpy(params->t35.payload_byte_data[idx], src->hdr_param->t35_param.payload_byte_data[idx], src->hdr_param->t35_param.payload_byte_length[idx]);
    }
    AddSideData(sd);
    sd->DecRef();
    side_data = m_frame->side_data[XMA_FRAME_SIDE_DATA_HDR10_PARAMS];
    sd_info   = static_cast<PrivateSideDataInfo*>(side_data->private_do_not_touch);
    if ((sd_info) && (sd_info->IsValid())) {
      sd_info->SetSelf(sd);
    }
  }

  // always add XMA_FRAME_SIDE_DATA_VUI_PARAMS
  sd = std::make_shared<InternalSideData>(m_sf, m_log, XMA_FRAME_SIDE_DATA_VUI_PARAMS, XMA_HOST_BUFFER_TYPE, sizeof(XmaVuiParams));
  if (!sd) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideData() Out of memory!");
  }
  memcpy(sd->GetHostPtr(), &src->vui_param, sizeof(XmaVuiParams));
  AddSideData(sd);
  sd->DecRef();
  side_data = m_frame->side_data[XMA_FRAME_SIDE_DATA_VUI_PARAMS];
  sd_info   = static_cast<PrivateSideDataInfo*>(side_data->private_do_not_touch);
  if ((sd_info) && (sd_info->IsValid())) {
    sd_info->SetSelf(sd);
  }

  // add XMA_FRAME_SIDE_DATA_ROI
  if (src->vpi_sideband[ROI_SB_INDEX]) {
    int            roi_size;
    unsigned char* roi_map;
    roi_size = src->vpi_sideband[ROI_SB_INDEX]->ch_info.width * src->vpi_sideband[ROI_SB_INDEX]->ch_info.height;
    sd       = std::make_shared<InternalSideData>(m_sf, m_log, XMA_FRAME_SIDE_DATA_ROI, XMA_HOST_BUFFER_TYPE, roi_size);
    if (!sd) {
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_FRAME, "InternalFrame::CopySideData() Out of memory!");
    }
    roi_map = reinterpret_cast<unsigned char*>(sd->GetHostPtr());
    memcpy(roi_map, src->vpi_sideband[ROI_SB_INDEX]->rc_data, roi_size);
    AddSideData(sd);
    sd->DecRef();
    side_data = m_frame->side_data[XMA_FRAME_SIDE_DATA_ROI];
    sd_info   = static_cast<PrivateSideDataInfo*>(side_data->private_do_not_touch);
    if ((sd_info) && (sd_info->IsValid())) {
      sd_info->SetSelf(sd);
    }
  }
}

///////////////////////////////////////////////////////////////////////////////
VpiBufferPool* GetVpiPool(LogPtr log, FramePtr frame) {
  if (!frame) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "GetVpiPool() Invalid parameter!");
  }

  FramePropertiesPtr props = frame->GetProps();

  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "GetVpiPool(format=%s, sw_format=%s, size=%dx%d)", InternalFrameProperties::GetPixFmtName(props->GetFormat()),
      InternalFrameProperties::GetPixFmtName(props->GetSwFormat()), props->GetWidth(), props->GetHeight());
  VpiBufferPool* pool      = nullptr;
  IntFramePtr    int_frame = std::dynamic_pointer_cast<InternalFrame>(frame);
  if (int_frame) {
    XmaFrame* xma_frame = int_frame->GetFrame();
    if (xma_frame) {
      LOCK_XMA_FRAME(log, "GetVpiPool", XMA_FRAME, xma_frame);
      PrivateFrameInfo* frame_info = static_cast<PrivateFrameInfo*>(xma_frame->private_do_not_touch);
      if (frame_info && frame_info->IsValid()) {
        pool = frame_info->GetBufferPool();
      }
    }
  }
  if (!pool) {
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "GetVpiPool() Unable to find buffer pool!");
  }
  return pool;
}

///////////////////////////////////////////////////////////////////////////////
int32_t AddFrameToVpiFrm(IntSessionFactoryPtr sf, LogPtr log, FramePtr frame, VpiFrm* pic) {
  if (!frame) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "AddFrameToVpiFrm() Invalid parameter!");
  }

  LOCK_XMA_FRAME(log, "AddFrameToVpiFrm", XMA_FRAME, std::dynamic_pointer_cast<InternalFrame>(frame)->m_frame);

  if (!pic) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "AddFrameToVpiFrm() Invalid parameter!");
  }

  FramePropertiesPtr props = frame->GetProps();

  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "AddFrameToVpiFrm(format=%s, sw_format=%s, size=%dx%d)", InternalFrameProperties::GetPixFmtName(props->GetFormat()),
      InternalFrameProperties::GetPixFmtName(props->GetSwFormat()), props->GetWidth(), props->GetHeight());

  if (((props->GetFormat() == XMA_VPE_FMT_TYPE) && (frame->GetBufferType() == XMA_HOST_BUFFER_TYPE)) ||
      ((props->GetFormat() != XMA_VPE_FMT_TYPE) && (frame->GetBufferType() != XMA_HOST_BUFFER_TYPE)) || (frame->GetBufferType() == NO_BUFFER)) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "AddFrameToVpiFrm() Inconsistent buffer types!");
  }

  int32_t planes = props->GetNumPlanes();
  int32_t plane;

  int32_t ch;
  if (!pic->valid) {
    pic->valid     = 1;
    pic->width     = props->GetWidth();
    pic->height    = props->GetHeight();
    pic->poc       = -1;
    pic->format    = static_cast<VpiFmt>(props->GetFormat());
    pic->key_frame = frame->IsIdr();
    pic->pts       = frame->GetPts();
    pic->dts       = -1;
    ch             = 0;
  } else {
    for (ch = 0; (ch < VPE_CHANNEL_NUM); ++ch) {
      if (!pic->ch_info[ch].enabled) {
        break;
      } else if ((props->GetWidth() == pic->ch_info[ch].width) && (props->GetHeight() == pic->ch_info[ch].height)) {
        THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME,
            "AddFrameToVpiFrm() Unable to add frame to pic, another frame with the same dimensions(%dx%d) has already been added!", props->GetWidth(),
            props->GetHeight());
      }
    }

    if (ch == VPE_CHANNEL_NUM) {
      THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "AddFrameToVpiFrm() No available free channels!");
    }
  }

  pic->ch_info[ch].enabled = 1;
  pic->ch_info[ch].width   = props->GetWidth();
  pic->ch_info[ch].height  = props->GetHeight();
  pic->ch_info[ch].walign  = XMA_FRAME_W_ALIGN;
  pic->ch_info[ch].halign  = XMA_FRAME_H_ALIGN;
  pic->ch_info[ch].flag    = props->GetFlags() | VPI_FRAME_NEW_FLAG;
  for (plane = 0; plane < planes; ++plane) {
    pic->ch_info[ch].stride_size[plane] = props->GetPlaneStride(plane);
  }

  pic->ch_info[ch].format    = static_cast<VpiFmt>(props->GetFormat());
  pic->ch_info[ch].sw_format = static_cast<VpiFmt>(props->GetSwFormat());

  switch (frame->GetBufferType()) {
  case XMA_HOST_BUFFER_TYPE:
  case XMA_DEVICE_ONLY_BUFFER_TYPE: {
  case XMA_DEVICE_BUFFER_TYPE:
    if (ch == 0) {
      buffer_t vpi_buf = sf->GetBuffer(0, XMA_HOST_BUFFER_TYPE);
      vpi_buffer_set_data(vpi_buf.get(), pic);
      vpi_buffer_set_free_cb(vpi_buf.get(), xma_pic_consume);
      vpi_buffer_set_opaque(vpi_buf.get(), nullptr);
      pic->vpi_picinfo_buf = vpi_buf.release();
    }

    PrivateVpiFrmInfo* info = static_cast<PrivateVpiFrmInfo*>(pic->opaque);
    if (info && info->IsValid()) {
      info->AddFrame(frame, pic, ch);
    } else {
      info = new PrivateVpiFrmInfo(sf, log);
      info->AddFrame(frame, pic, ch);
      VpiBuffer* vpi_buf = static_cast<VpiBuffer*>(pic->vpi_picinfo_buf);
      info->SetFreeBuf(vpi_buf->freebuf);
      vpi_buf->freebuf = xma_free_reference_info;
      info->SetPool(static_cast<VpiBufferPool*>(vpi_buf->pool));
      vpi_buf->pool = GetVpiPool(log, frame);
      pic->opaque   = info;
    }
    vpi_frame_ref(pic);
  } break;
  case NO_BUFFER:
    break;
  default:
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "AddFrameToVpiFrm() Unknown buffer type!");
  }

  AddSideDataToVpiFrm(log, frame, pic);

  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "AddFrameToVpiFrm() - Returning");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t CopyFrame(LogPtr log, XmaFrame* dst, FramePtr src) {
  if (!dst || (dst->data[0].buffer_type != NO_BUFFER) || !src) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "CopyFrame() Invalid parameter!");
  }

  LOCK_XMA_FRAME(log, "CopyFrame", XMA_FRAME, std::dynamic_pointer_cast<InternalFrame>(src)->m_frame);

  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "CopyFrame(format=%s, sw_format=%s, size=%dx%d)", InternalFrameProperties::GetPixFmtName(src->GetProps()->GetFormat()),
      InternalFrameProperties::GetPixFmtName(src->GetProps()->GetSwFormat()), src->GetProps()->GetWidth(), src->GetProps()->GetHeight());

  IntFramePtr src_frame = std::dynamic_pointer_cast<InternalFrame>(src);
  if (!src_frame) {
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "CopyFrame() source not valid!");
  }

  PrivateFrameInfo* dst_info = static_cast<PrivateFrameInfo*>(dst->private_do_not_touch);
  if ((!dst_info) || (!dst_info->IsValid())) {
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "CopyFrame() Invalid private info!");
  }

  memcpy(dst, src_frame->m_frame.get(), sizeof(XmaFrame));

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(src_frame->m_sf->m_mutex);
  src_frame->m_sf->m_frames.remove(src_frame->m_frame.get());
  lock.unlock();
#endif

  IntFramePtr dst_frame = dst_info->GetSelf();
  src_frame->m_frame    = std::move(dst_frame->m_frame);
  delete dst_info;

  PrivateFrameInfo* src_info = static_cast<PrivateFrameInfo*>(src_frame->m_frame->private_do_not_touch);
  if ((!src_info) || (!src_info->IsValid())) {
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_FRAME, "CopyFrame() Invalid private info!");
  }
  src_info->SetSelf(src_frame);

  log->LogMsg(XMA_DEBUG_LOG, XMA_FRAME, "CopyFrame() - Returning");
  return XMA_SUCCESS;
}

}; // Namespace xma
