/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#include "xmaapi.h"
#include <stdarg.h>
#include <time.h>
#include <cstring>
#include <iostream>
#include "chrono"
#include "thread"

namespace xma {

#define XMA_LOGGER_MOD "xmalogger"

///////////////////////////////////////////////////////////////////////////////
std::string timestamp() {
  auto        time    = std::chrono::system_clock::now();
  std::time_t t       = std::chrono::system_clock::to_time_t(time);
  std::tm*    tm      = gmtime(&t);
  char        buf[64] = {0};
  return std::strftime(buf, sizeof(buf), "%c GMT", tm) ? buf : "Time conversion failed";
}

///////////////////////////////////////////////////////////////////////////////
LogPtr Log::MakeLog(XmaLogLevelType log_level, XmaLogType log_type, ...) {
  va_list ap;
  va_start(ap, log_type);

  LogPtr log;
  switch (log_type) {
  case XMA_LOG_TYPE_NONE:
    log = std::make_shared<XmaLogNull>(log_level);
    break;
  case XMA_LOG_TYPE_CONSOLE:
    log = std::make_shared<XmaLogConsole>(log_level);
    break;
  case XMA_LOG_TYPE_SYSLOG:
    log = std::make_shared<XmaLogSyslog>(log_level);
    break;
  case XMA_LOG_TYPE_FILE: {
    const char* filename = va_arg(ap, const char*);
    log                  = std::make_shared<XmaLogFile>(log_level, filename);
    break;
  }
  case XMA_LOG_TYPE_CALLBACK: {
    XmaLogCallbackFunction callback = va_arg(ap, XmaLogCallbackFunction);
    void*                  opaque   = va_arg(ap, void*);
    log                             = std::make_shared<XmaLogCallback>(log_level, callback, opaque);
    break;
  }
  default:
    throw InvalidArgumentException("Log::MakeLogList() Invalid log type!");
  }

  va_end(ap);

  return log;
}

///////////////////////////////////////////////////////////////////////////////
LogPtr Log::MakeLogList(XmaLogLevelType log_level, XmaLogType log_type, va_list arg) {
  switch (log_type) {
  case XMA_LOG_TYPE_NONE:
    return std::make_shared<XmaLogNull>(log_level);
  case XMA_LOG_TYPE_CONSOLE:
    return std::make_shared<XmaLogConsole>(log_level);
  case XMA_LOG_TYPE_SYSLOG:
    return std::make_shared<XmaLogSyslog>(log_level);
  case XMA_LOG_TYPE_FILE: {
    const char* filename = va_arg(arg, const char*);
    return std::make_shared<XmaLogFile>(log_level, filename);
  }
  case XMA_LOG_TYPE_CALLBACK: {
    XmaLogCallbackFunction callback = va_arg(arg, XmaLogCallbackFunction);
    void*                  opaque   = va_arg(arg, void*);
    return std::make_shared<XmaLogCallback>(log_level, callback, opaque);
  }
  default:
    throw InvalidArgumentException("Log::MakeLogList() Invalid log type!");
  }
}

///////////////////////////////////////////////////////////////////////////////
void Log::LogMsg(XmaLogLevelType level, const char* name, const char* msg, ...) {
  va_list ap;
  va_start(ap, msg);

  LogMsgList(level, name, msg, ap);

  va_end(ap);
}

///////////////////////////////////////////////////////////////////////////////
void XmaLog::LogMsgList(XmaLogLevelType level, const char* name, const char* msg, va_list arg) {
  if (level <= m_level) {
    /* Create message buffer on the stack */
    char msg_buff[XMA_MAX_LOGMSG_SIZE];
    memset(msg_buff, 0, sizeof(msg_buff));

    /* Set component name */
    char log_name[40] = {0};
    if (name == nullptr) {
      strncpy(log_name, "XMA-default", sizeof(log_name));
    } else {
      strncpy(log_name, name, sizeof(log_name) - 1);
    }

    snprintf(msg_buff, XMA_MAX_LOGMSG_SIZE, "%s %s ", program_invocation_short_name, log_name);
    int32_t hdr_offset = strlen(msg_buff);
    vsnprintf(&msg_buff[hdr_offset], (XMA_MAX_LOGMSG_SIZE - hdr_offset), msg, arg);

    std::lock_guard<std::mutex> guard(m_mutex);

    Log(level, "XMA", msg_buff);
  }
}

///////////////////////////////////////////////////////////////////////////////
XmaLogLevelType XmaLog::GetLogLevel() const {
  return m_level;
}

///////////////////////////////////////////////////////////////////////////////
const char* XmaLog::GetFileName() const {
  throw RuntimeException("XmaLog::GetFileName() XmaLogType is not XMA_LOG_TYPE_FILE!");
}

///////////////////////////////////////////////////////////////////////////////
void XmaLogConsole::Log(XmaLogLevelType level, const char* tag, const char* msg) {
  std::cerr << "[" << tag << "] " << severityMap[level] << msg << std::endl;
}

///////////////////////////////////////////////////////////////////////////////
void XmaLogSyslog::Log(XmaLogLevelType level, const char*, const char* msg) {
  syslog(severityMap[level], "%s", msg);
}

///////////////////////////////////////////////////////////////////////////////
XmaLogFile::XmaLogFile(XmaLogLevelType level, const char* filename) : XmaLog(level) {
  m_handle.open(filename, std::ofstream::out | std::ofstream::app);
}

///////////////////////////////////////////////////////////////////////////////
XmaLogFile::~XmaLogFile() {
  m_handle.close();
}

///////////////////////////////////////////////////////////////////////////////
void XmaLogFile::Log(XmaLogLevelType level, const char* tag, const char* msg) {
  m_handle << "[" << timestamp() << "] [" << tag << "] Tid: " << std::this_thread::get_id() << ", " << severityMap[level] << msg << std::endl;
}

///////////////////////////////////////////////////////////////////////////////
void XmaLogCallback::LogMsgList(XmaLogLevelType level, const char* name, const char* msg, va_list arg) {
  if (level <= m_level) {
    /* Create message buffer on the stack */
    char msg_buff[XMA_MAX_LOGMSG_SIZE];
    memset(msg_buff, 0, sizeof(msg_buff));

    /* Set component name */
    char log_name[40] = {0};
    if (name == nullptr) {
      strncpy(log_name, "XMA-default", sizeof(log_name));
    } else {
      strncpy(log_name, name, sizeof(log_name) - 1);
    }

    vsnprintf(msg_buff, XMA_MAX_LOGMSG_SIZE, msg, arg);

    std::lock_guard<std::mutex> guard(m_mutex);

    Log(level, name, msg_buff);
  }
}

///////////////////////////////////////////////////////////////////////////////
void XmaLogCallback::Log(XmaLogLevelType level, const char* tag, const char* msg) {
  m_callback(m_opaque, level, tag, msg);
}

} // namespace xma
