/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#pragma once

#include <memory>
#include <mutex>
#include "xmacpp/DataBuffer.h"
#include "InternalSessionFactory.h"

namespace xma {

int32_t CopyDataBuffer(LogPtr log, XmaDataBuffer* dst, DataBufferPtr src);

//////////////////////////////////////////////////////////////////////////////
class InternalDataBuffer final : public DataBuffer {
public:
  InternalDataBuffer(IntSessionFactoryPtr sf, LogPtr log, size_t size, bool dummy, int64_t pts = 0, int64_t poc = 0);
  InternalDataBuffer(
      IntSessionFactoryPtr sf, LogPtr log, uint8_t* data, size_t size, std::function<void(uint8_t*)> free_callback = NULL, int64_t pts = 0, int64_t poc = 0);
  virtual ~InternalDataBuffer();

  XmaDataBuffer* GetDataBuffer();
  int32_t        IncRef();
  int32_t        DecRef();
  int32_t        GetRefCount() const;
  void           SetSelf();

  virtual uint8_t* GetBuffer() const override;
  virtual int32_t  GetSize() const override;
  virtual int64_t  GetPts() const override;
  virtual int64_t  GetPoc() const override;

  int32_t SetSize(int32_t size);
  int32_t SetPts(int64_t pts);
  int32_t SetPoc(int64_t poc);

  friend int32_t CopyDataBuffer(LogPtr log, XmaDataBuffer* dst, DataBufferPtr src);

protected:
  std::unique_ptr<XmaDataBuffer> m_buf;
  IntSessionFactoryPtr           m_sf;
  LogPtr                         m_log;
};

using IntDataBufferPtr = std::shared_ptr<InternalDataBuffer>;

//////////////////////////////////////////////////////////////////////////////
#define DATA_BUFFER_DUMMY_VALUE 0xFEDCBA
class PrivateDataBufferInfo {
public:
  PrivateDataBufferInfo(IntSessionFactoryPtr sf, LogPtr log);
  ~PrivateDataBufferInfo();

  bool                  IsValid() const;
  std::recursive_mutex& GetMutex();
  void                  SetSelf(DataBufferPtr buf);
  DataBufferPtr         GetSelf() const;
  void                  SetCallback(std::function<void(uint8_t*)> free_callback);
  void                  CallCallback(uint8_t* data);
  IntSessionFactoryPtr  GetSessionFactory();
  LogPtr                GetLog();

private:
  uint32_t                      m_dummy = DATA_BUFFER_DUMMY_VALUE;
  mutable std::recursive_mutex  m_mutex;
  DataBufferPtr                 m_self; // C implementation uses this
  std::function<void(uint8_t*)> m_free_callback = nullptr;
  IntSessionFactoryPtr          m_sf;
  LogPtr                        m_log;
};

#define LOCK_DATA_BUFFER(log, func, module, data_buffer) \
  if (!data_buffer) { \
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, module, "%s() XmaDataBuffer is not valid!", func); \
  } \
  xma::PrivateDataBufferInfo* private_data_buffer_info = static_cast<xma::PrivateDataBufferInfo*>(data_buffer->private_do_not_touch); \
  if ((!private_data_buffer_info) || (!private_data_buffer_info->IsValid())) { \
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, module, "%s() Private info is not valid!", func); \
  } \
  std::unique_lock<std::recursive_mutex> guard(private_data_buffer_info->GetMutex());

}; // Namespace xma
