/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#pragma once

#include <memory>
#include "xmacpp/SideData.h"
#include "InternalSessionFactory.h"
#include "BufferPool.h"
#include "vpi_memory.h"

namespace xma {
//////////////////////////////////////////////////////////////////////////////
class InternalSideData final : public SideData {
public:
  InternalSideData(IntSessionFactoryPtr sf, LogPtr log, enum XmaFrameSideDataType type, XmaBufferType buffer_type, uint64_t size);
  virtual ~InternalSideData();

  XmaFrameSideData* GetSideData();
  int32_t           IncRef();
  int32_t           DecRef();
  int32_t           GetRefCount() const;

  virtual XmaFrameSideDataType GetType() const override;
  virtual XmaBufferType        GetBufferType() const override;
  virtual bool                 IsOnHost() const override;
  virtual uint8_t*             GetHostPtr() const override;
  virtual uint64_t             GetDevicePtr() const override;
  virtual size_t               GetSize() const override;

  virtual int32_t Read() override;
  virtual int32_t Write() override;

protected:
  std::unique_ptr<XmaFrameSideData> m_sd;
  IntSessionFactoryPtr              m_sf;
  LogPtr                            m_log;
};

using IntSideDataPtr = std::shared_ptr<InternalSideData>;

int32_t AddSideDataToVpiFrm(LogPtr log, FramePtr src, VpiFrm* dst);

//////////////////////////////////////////////////////////////////////////////
#define SIDE_DATA_DUMMY_VALUE 0xCDCDCDCD
class PrivateSideDataInfo {
public:
  PrivateSideDataInfo(IntSessionFactoryPtr sf, LogPtr log, buffer_t buffer);
  ~PrivateSideDataInfo();

  bool                  IsValid() const;
  std::recursive_mutex& GetMutex();
  VpiLinearMem*         GetMem() const;
  void                  SetSelf(IntSideDataPtr self);
  IntSideDataPtr        GetSelf() const;
  void                  SetSelfPtr(InternalSideData* self_ptr);
  InternalSideData*     GetSelfPtr() const;
  IntSessionFactoryPtr  GetSessionFactory() const;
  LogPtr                GetLog() const;

private:
  uint32_t                     m_dummy = SIDE_DATA_DUMMY_VALUE;
  mutable std::recursive_mutex m_mutex;
  buffer_t                     m_buffer;
  IntSideDataPtr               m_self;     // C implementation uses this
  InternalSideData*            m_self_ptr; // C++ implementation uses this
  IntSessionFactoryPtr         m_sf;
  LogPtr                       m_log;
};

#define LOCK_SIDE_DATA(log, func, module, side_data) \
  if (!side_data) { \
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, module, "%s() XmaFrameSideData is not valid!", func); \
  } \
  xma::PrivateSideDataInfo* private_side_data_info = static_cast<xma::PrivateSideDataInfo*>(side_data->private_do_not_touch); \
  if ((!private_side_data_info) || (!private_side_data_info->IsValid())) { \
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, module, "%s() Private info is not valid!", func); \
  } \
  std::unique_lock<std::recursive_mutex> guard(private_side_data_info->GetMutex());

}; // Namespace xma
