/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#include <cstring>
#include "xmaapiencoder.h"
#include "xmaapi.h"
#include "InternalFrame.h"
#include "InternalDataBuffer.h"

#define XMA_ENCODER_MOD "xmaencoder"

///////////////////////////////////////////////////////////////////////////////
XmaEncoderSession* xma_enc_session_create(XmaEncoderProperties* enc_props) {
  if (!enc_props) {
    return nullptr;
  }

  XmaSessionFactory* xma_sf = static_cast<XmaSessionFactory*>(enc_props->handle);
  if (!xma_sf) {
    return nullptr;
  }

  TRY()
  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_create()");
  XmaEncoderSession* session = new XmaEncoderSession();
  if (!session) {
    xma_sf->m_sf->GetLog()->LogMsg(XMA_ERROR_LOG, XMA_ENCODER_MOD, "xma_enc_session_create() Out Of Memory!");
    return nullptr;
  }

  session->m_sf  = xma_sf->m_sf;
  session->m_log = xma_sf->m_sf->GetLog();

  session->m_session = session->m_sf->MakeEncoderSession(enc_props);
  if (!session->m_session) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_ENCODER_MOD, "xma_enc_session_create() Out Of Memory!");
    delete session;
    return nullptr;
  }

  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_create() - Returning");
  return static_cast<XmaEncoderSession*>(session);
  CATCH_NULL(xma_sf->m_sf->GetLog(), XMA_ENCODER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_enc_session_destroy(XmaEncoderSession* session) {
  if (!session) {
    return XMA_ERROR_INVALID;
  }

  xma::LogPtr log = session->m_log;
  log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_destroy()");

  if (!session->m_session) {
    log->LogMsg(XMA_ERROR_LOG, XMA_ENCODER_MOD, "xma_enc_session_destroy() Invalid parameter!");
    delete session;
    return XMA_ERROR_INVALID;
  }

  TRY()
  delete session;

  log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_destroy() - Returning success");
  return XMA_SUCCESS;
  CATCH_ERROR(log, XMA_ENCODER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_enc_session_set_log(XmaEncoderSession* session, XmaLogHandle handle) {
  if (!session || !handle) {
    return XMA_ERROR_INVALID;
  }

  TRY()
  XmaLogStruct* log_struct = static_cast<XmaLogStruct*>(handle);
  xma::LogPtr   log        = log_struct->m_log;
  const char*   level;
  switch (log->GetLogLevel()) {
  case XMA_EMERGENCY_LOG:
    level = "emergency";
    break;
  case XMA_ALERT_LOG:
    level = "alert";
    break;
  case XMA_CRITICAL_LOG:
    level = "critical";
    break;
  case XMA_ERROR_LOG:
    level = "error";
    break;
  case XMA_WARNING_LOG:
    level = "warning";
    break;
  case XMA_NOTICE_LOG:
    level = "notice";
    break;
  case XMA_INFO_LOG:
    level = "info";
    break;
  case XMA_DEBUG_LOG:
    level = "debug";
    break;
  default:
    level = "unknown";
    break;
  }
  switch (log->GetLogType()) {
  case XMA_LOG_TYPE_NONE:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_set_log(level=%s, type=none)", level);
    break;
  case XMA_LOG_TYPE_CONSOLE:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_set_log(level=%s, type=console)", level);
    break;
  case XMA_LOG_TYPE_SYSLOG:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_set_log(level=%s, type=syslog)", level);
    break;
  case XMA_LOG_TYPE_FILE:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_set_log(level=%s, type=file, file='%s')", level, log->GetFileName());
    break;
  case XMA_LOG_TYPE_CALLBACK:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_set_log(level=%s, type=callback)", level);
    break;
  default:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_set_log(level=%s, type=unknown)", level);
    break;
  }
  session->m_log = log;
  session->m_session->SetLog(log);

  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_set_log() - Returning success!");
  return XMA_SUCCESS;
  CATCH_ERROR(session->m_log, XMA_ENCODER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_enc_session_send_frame(XmaEncoderSession* session, XmaFrame* xma_frame) {
  if (!session) {
    return XMA_ERROR_INVALID;
  }

  if (!session->m_session) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_ENCODER_MOD, "xma_enc_session_send_frame() Invalid parameter!");
    return XMA_ERROR_INVALID;
  }

  TRY()
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_send_frame()");

  xma::queue_t queue = xma::QueueFrame();
  if (xma_frame) {
    xma::PrivateFrameInfo* info = static_cast<xma::PrivateFrameInfo*>(xma_frame->private_do_not_touch);
    xma::IntFramePtr       frame;
    if ((info) && (info->IsValid())) {
      frame = info->GetSelf();
    }
    if (frame) {
      std::get<xma::QueueFrame>(queue).frame = frame;
      if (frame->GetRefCount() == 1) {
        info->SetSelf(nullptr);
      } else {
        frame->DecRef();
      }
    }
  }
  int32_t ret = session->m_session->Enqueue(queue);
  char    error_msg[256];
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_send_frame() - Returning %s", xma_perror(ret, error_msg, sizeof(error_msg)));
  return ret;
  CATCH_ERROR(session->m_log, XMA_ENCODER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_enc_session_recv_data(XmaEncoderSession* session, XmaDataBuffer* data, int32_t* data_size) {
  if (!session) {
    return XMA_ERROR_INVALID;
  }

  if (!session->m_session) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_ENCODER_MOD, "xma_enc_session_recv_data() Invalid parameter!");
    return XMA_ERROR_INVALID;
  }

  if (!data || (data->data.buffer_type != NO_BUFFER)) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_ENCODER_MOD, "xma_enc_session_recv_data() Invalid parameter!");
    return XMA_ERROR_INVALID;
  }

  TRY()
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_recv_data()");
  xma::queue_t queue;
  int32_t      ret = session->m_session->Dequeue(queue);
  if (ret != XMA_SUCCESS) {
    char error_msg[256];
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_recv_data() - Returning %s", xma_perror(ret, error_msg, sizeof(error_msg)));
    return ret;
  }

  CopyDataBuffer(session->m_log, data, std::move(std::get<xma::QueueDataBuffer>(queue).data));
  if (data_size) {
    *data_size = data->alloc_size;
  }

  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_MOD, "xma_enc_session_recv_data() - Returning success!");
  return ret;
  CATCH_ERROR(session->m_log, XMA_ENCODER_MOD)
}
