/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#include <cstring>
#include "xmaapiscaler.h"
#include "xmaapi.h"
#include "InternalFrame.h"

#define XMA_SCALER_MOD "xmascaler"

///////////////////////////////////////////////////////////////////////////////
XmaScalerSession* xma_scaler_session_create(XmaScalerProperties* props) {
  if (!props) {
    return nullptr;
  }

  XmaSessionFactory* xma_sf = static_cast<XmaSessionFactory*>(props->handle);
  if (!xma_sf) {
    return nullptr;
  }

  TRY()
  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_create()");
  XmaScalerSession* session = new XmaScalerSession();
  if (!session) {
    xma_sf->m_sf->GetLog()->LogMsg(XMA_ERROR_LOG, XMA_SCALER_MOD, "xma_scaler_session_create() Out Of Memory!");
    return nullptr;
  }

  session->m_sf     = xma_sf->m_sf;
  session->m_log    = xma_sf->m_sf->GetLog();
  session->m_handle = props->handle;

  session->m_session = session->m_sf->MakeScalerSession(props);
  if (!session->m_session) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_SCALER_MOD, "xma_scaler_session_create() Out Of Memory!");
    delete session;
    return nullptr;
  }

  session->m_outputs = props->num_outputs;

  memset(session->m_props, 0, sizeof(XmaFrameProperties) * MAX_SCALER_OUTPUTS);
  for (int32_t output = 0; output < session->m_outputs; ++output) {
    session->m_props[output].format    = props->output[output].format;
    session->m_props[output].sw_format = props->output[output].sw_format;
    session->m_props[output].width     = props->output[output].width;
    session->m_props[output].height    = props->output[output].height;
    int32_t planes                     = xma_frame_planes_get(props->handle, &session->m_props[output]);
    for (int32_t plane = 0; plane < planes; ++plane) {
      session->m_props[output].linesize[plane] = xma_frame_get_plane_stride(props->handle, &session->m_props[output], plane);
    }
  }

  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_create() - Returning");
  return static_cast<XmaScalerSession*>(session);
  CATCH_NULL(xma_sf->m_sf->GetLog(), XMA_SCALER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_scaler_session_destroy(XmaScalerSession* session) {
  if (!session) {
    return XMA_ERROR_INVALID;
  }

  xma::LogPtr log = session->m_log;
  log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_destroy()");

  if (!session->m_session) {
    log->LogMsg(XMA_ERROR_LOG, XMA_SCALER_MOD, "xma_scaler_session_destroy() Invalid parameter!");
    delete session;
    return XMA_ERROR_INVALID;
  }

  TRY()
  delete session;

  log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_destroy() - Returning success");
  return XMA_SUCCESS;
  CATCH_ERROR(log, XMA_SCALER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_scaler_session_set_log(XmaScalerSession* session, XmaLogHandle handle) {
  if (!session || !handle) {
    return XMA_ERROR_INVALID;
  }

  TRY()
  XmaLogStruct* log_struct = static_cast<XmaLogStruct*>(handle);
  xma::LogPtr   log        = log_struct->m_log;
  const char*   level;
  switch (log->GetLogLevel()) {
  case XMA_EMERGENCY_LOG:
    level = "emergency";
    break;
  case XMA_ALERT_LOG:
    level = "alert";
    break;
  case XMA_CRITICAL_LOG:
    level = "critical";
    break;
  case XMA_ERROR_LOG:
    level = "error";
    break;
  case XMA_WARNING_LOG:
    level = "warning";
    break;
  case XMA_NOTICE_LOG:
    level = "notice";
    break;
  case XMA_INFO_LOG:
    level = "info";
    break;
  case XMA_DEBUG_LOG:
    level = "debug";
    break;
  default:
    level = "unknown";
    break;
  }
  switch (log->GetLogType()) {
  case XMA_LOG_TYPE_NONE:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_set_log(level=%s, type=none)", level);
    break;
  case XMA_LOG_TYPE_CONSOLE:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_set_log(level=%s, type=console)", level);
    break;
  case XMA_LOG_TYPE_SYSLOG:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_set_log(level=%s, type=syslog)", level);
    break;
  case XMA_LOG_TYPE_FILE:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_set_log(level=%s, type=file, file='%s')", level, log->GetFileName());
    break;
  case XMA_LOG_TYPE_CALLBACK:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_set_log(level=%s, type=callback)", level);
    break;
  default:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_set_log(level=%s, type=unknown)", level);
    break;
  }
  session->m_log = log;
  session->m_session->SetLog(log);

  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_set_log() - Returning success!");
  return XMA_SUCCESS;
  CATCH_ERROR(session->m_log, XMA_SCALER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_scaler_session_send_frame(XmaScalerSession* session, XmaFrame* xma_frame) {
  if (!session) {
    return XMA_ERROR_INVALID;
  }

  if (!session->m_session) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_SCALER_MOD, "xma_scaler_session_send_frame() Invalid parameter!");
    return XMA_ERROR_INVALID;
  }

  TRY()
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_send_frame()");

  int32_t ret = XMA_SUCCESS;
  if (!xma_frame) {
    xma::queue_t queue = xma::QueueFrame();
    ret                = session->m_session->Enqueue(queue);
  } else {
    xma::PrivateFrameInfo* info = static_cast<xma::PrivateFrameInfo*>(xma_frame->private_do_not_touch);
    if ((!info) || (!info->IsValid())) {
      goto fail;
    }
    xma::IntFramePtr frame = info->GetSelf();
    if (!frame) {
      goto fail;
    }
    xma::queue_t queue                     = xma::QueueFrame();
    std::get<xma::QueueFrame>(queue).frame = frame;
    ret                                    = session->m_session->Enqueue(queue);
    if (ret != XMA_TRY_AGAIN) {
      if (frame->GetRefCount() == 1) {
        info->SetSelf(nullptr);
      } else {
        frame->DecRef();
      }
    }
  }
  char error_msg[256];
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_send_frame() - Returning %s", xma_perror(ret, error_msg, sizeof(error_msg)));
  return ret;
  CATCH_ERROR(session->m_log, XMA_SCALER_MOD)
fail:
  session->m_log->LogMsg(XMA_ERROR_LOG, XMA_SCALER_MOD, "xma_scaler_session_send_frame() Invalid frame!");
  return XMA_ERROR_INVALID;
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_scaler_session_recv_frame_list(XmaScalerSession* session, XmaFrame** frame_list) {
  if (!session) {
    return XMA_ERROR_INVALID;
  }

  if (!session->m_session) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_SCALER_MOD, "xma_scaler_session_recv_frame_list() Invalid parameter!");
    return XMA_ERROR_INVALID;
  }

  if (!frame_list) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_SCALER_MOD, "xma_scaler_session_recv_frame_list() Invalid parameter!");
    return XMA_ERROR_INVALID;
  }

  TRY()
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_recv_frame_list()");

  xma::queue_t queue;
  int32_t      ret = session->m_session->Dequeue(queue);
  if (ret != XMA_SUCCESS) {
    char error_msg[256];
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_recv_frame_list() - Returning %s", xma_perror(ret, error_msg, sizeof(error_msg)));
    return ret;
  }

  for (int32_t output = 0; output < session->m_outputs; ++output) {
    if (!frame_list[output]) {
      frame_list[output] = xma_frame_alloc(session->m_handle, &session->m_props[output], true);
    }

    CopyFrame(session->m_log, frame_list[output], std::move(std::get<xma::QueueFrameList>(queue).frame_list[output]));
  }

  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_SCALER_MOD, "xma_scaler_session_recv_frame_list() - Returning success!");
  return ret;
  CATCH_ERROR(session->m_log, XMA_SCALER_MOD)
}
